#define _DEFAULT_SOURCE /* for usleep() */
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>

#include <time.h>
#include <unistd.h>

#include <string.h>
#include <errno.h>

#include "chip8.h"

/* instructions per frame */
#define SPEED 2000

/* stack size in levels */
#define STACK 12

/* instruction trace */
#if 0
#define TRACE(...) printf(__VA_ARGS__)
#else
#define TRACE(...)
#endif


/* emulator state */
struct emulator {
	struct chip8_callbacks cb;
	uint8_t vram[8*32];
	union {
		uint8_t mem[4096];
		struct {
			uint8_t  font[80]; /* Preloaded Font   */
			uint16_t pc;       /* Program Counter  */
			uint16_t i;        /* Address Register */
			uint8_t  v[16];    /* Data Registers   */
			uint16_t s[STACK]; /* Stack Space      */
			uint8_t  sp;       /* Stack Pointer    */
			uint8_t  dtim;     /* Delay Timer      */
			uint8_t  stim;     /* Sound Timer      */
			uint16_t seed;     /* Random Seed      */
		};
	};
};

/* built-in font */
const uint8_t font[80] = {
	0xF0, 0x90, 0x90, 0x90, 0xF0, /* 0 */
	0x20, 0x60, 0x20, 0x20, 0x70, /* 1 */
	0xF0, 0x10, 0xF0, 0x80, 0xF0, /* 2 */
	0xF0, 0x10, 0xF0, 0x10, 0xF0, /* 3 */
	0x90, 0x90, 0xF0, 0x10, 0x10, /* 4 */
	0xF0, 0x80, 0xF0, 0x10, 0xF0, /* 5 */
	0xF0, 0x80, 0xF0, 0x90, 0xF0, /* 6 */
	0xF0, 0x10, 0x20, 0x40, 0x40, /* 7 */
	0xF0, 0x90, 0xF0, 0x90, 0xF0, /* 8 */
	0xF0, 0x90, 0xF0, 0x10, 0xF0, /* 9 */
	0xF0, 0x90, 0xF0, 0x90, 0x90, /* A */
	0xE0, 0x90, 0xE0, 0x90, 0xE0, /* B */
	0xF0, 0x80, 0x80, 0x80, 0xF0, /* C */
	0xE0, 0x90, 0x90, 0x90, 0xE0, /* D */
	0xF0, 0x80, 0xF0, 0x80, 0xF0, /* E */
	0xF0, 0x80, 0xF0, 0x80, 0x80, /* F */
};

/* create and initialize new CHIP-8 object */
void* chip8_create(struct chip8_callbacks* cb)
{
	/* check if callbacks are present */
	if ((cb->tick    == NULL) ||
	    (cb->draw    == NULL) ||
	    (cb->getkey  == NULL) ||
	    (cb->waitkey == NULL)) {
		fprintf(stderr, "CHIP-8: required callback(s) missing\n");
		return NULL;
	}

	/* allocate emulator object */
	struct emulator* new = calloc(1, sizeof(struct emulator));
	if (!new) {
		fprintf(stderr, "CHIP-8: failed to allocate\n");
		return NULL;
	}

	/* save emulator callbacks */
	new->cb = *cb;

	/* initialize emulator state */
	memcpy(new->font, font, sizeof(font));
	new->seed = time(0);
	new->pc = 0x200;

	return new;
}

/* destroy CHIP-8 object */
void chip8_destroy(void* handle)
{
	struct emulator* emu = (struct emulator*)handle;

	/* stop sound if playing */
	if (emu->cb.sound)
		emu->cb.sound(0);

	free(emu);
}

/* load a binary file into memory */
int chip8_load(void* handle, const char* file)
{
	struct emulator* emu = (struct emulator*)handle;

	FILE* fh = fopen(file, "rb");
	if (!fh) {
		fprintf(stderr, "CHIP-8: open failed: %s\n", strerror(errno));
		return 1;
	}

	/* read at most 3584 bytes */
	size_t ret = fread(&emu->mem[0x200], 1, 3584, fh);
	if (ret == 0 || (ret < 3584 && !feof(fh))) {
		fprintf(stderr, "CHIP-8: load failed: %s\n", strerror(errno));
		return 1;
	}

	fclose(fh);
	return 0;
}

/* draw a sprite */
static int draw_sprite(struct emulator* emu, int x, int y, int bytes)
{
	TRACE("DRAW AT (%d,%d) FROM %03x BYTES ", x, y, emu->i);
	for (int i = 0; i < bytes; i++)
		TRACE("%02x ", emu->mem[(emu->i + i) & 0xFFF]);
	TRACE("\n");

	int ret = 0;
	for (int yy = y; yy < y + bytes; yy++) {
		uint8_t sdata = emu->mem[(emu->i+yy-y) & 0xFFF];
		for (int xx = x; xx < x + 8; xx++) {
			/* extract current and new pixel value */
			int pixel = yy * 64 + xx;
			int voff  = pixel / 8;
			int vmask = 1 << (7 - pixel % 8);
			int old = !!(emu->vram[voff] & vmask);
			int new = old ^ !!(sdata & 0x80);
			sdata <<= 1;

			/* remember any collision */
			if (old && !new) ret = 1;

			/* update pixel */
			if (new) {
				emu->vram[voff] |= vmask;
				emu->cb.draw(xx, yy, 1);
			} else {
				emu->vram[voff] &= ~vmask;
				emu->cb.draw(xx, yy, 0);
			}
		}
	}

	return ret;
}

/* emulate a single instruction */
static int next_instr(struct emulator* emu)
{
	/* read next instruction */
	TRACE("PC:%03x ", emu->pc);
	uint16_t instr =
		emu->mem[(emu->pc)   & 0xFFF] << 8 |
		emu->mem[(emu->pc+1) & 0xFFF];
	emu->pc = (emu->pc + 2) & 0xFFF;

	/* extract bitfields */
	int op  = (instr & 0xF000) >> 12;
	int nnn = (instr & 0x0FFF);
	int nn  = (instr & 0x00FF);
	int n   = (instr & 0x000F);
	int y   = (instr & 0x00F0) >> 4;
	int x   = (instr & 0x0F00) >> 8;

	/* dump state */
	TRACE("INSTR:%04x I:%03x SP:%02x V:"
		"%02x%02x%02x%02x %02x%02x%02x%02x "
		"%02x%02x%02x%02x %02x%02x%02x%02x "
		"DT:%02x ST:%02x\n",
		instr, emu->i, emu->sp,
		emu->v[0x0], emu->v[0x1], emu->v[0x2], emu->v[0x3],
		emu->v[0x4], emu->v[0x5], emu->v[0x6], emu->v[0x7],
		emu->v[0x8], emu->v[0x9], emu->v[0xA], emu->v[0xB],
		emu->v[0xC], emu->v[0xD], emu->v[0xE], emu->v[0xF],
		emu->dtim, emu->stim);

	/* instruction emulation */
	if (op == 0x0 && nn == 0xE0) {
		/* clear screen */
		memset(emu->vram, 0, sizeof(emu->vram));
		for (int yy = 0; yy < 32; yy++)
			for (int xx = 0; xx < 64; xx++)
				emu->cb.draw(xx, yy, 0);

	} else if (op == 0x0 && nn == 0xEE) {
		/* return from subroutine */
		if (!emu->sp) {
			fprintf(stderr, "CHIP-8: return on empty at %04x\n",
				(emu->pc-2) & 0xFFF);
			return 1; /* EXIT: return on empty stack */
		}

		emu->sp--;
		emu->pc = emu->s[emu->sp];

	} else if (op == 0x1) {
		/* jump to nnn */
		emu->pc = nnn;

	} else if (op == 0x2) {
		/* execute subroutine at nnn */
		if (emu->sp == STACK-1) {
			fprintf(stderr, "CHIP-8: stack overflow at %04x\n",
				(emu->pc-2) & 0xFFF);
			return 3; /* EXIT: stack overflow */
		}

		emu->s[emu->sp] = emu->pc;
		emu->sp++;
		emu->pc = nnn;

	} else if (op == 0x3) {
		/* skip if v[x] == nn */
		if (emu->v[x] == nn)
			emu->pc += 2;

	} else if (op == 0x4) {
		/* skip if v[x] != nn */
		if (emu->v[x] != nn)
			emu->pc += 2;

	} else if (op == 0x5) {
		/* skip if v[x] == v[y] */
		if (emu->v[x] == emu->v[y])
			emu->pc += 2;

	} else if (op == 0x6) {
		/* store nn in v[x] */
		emu->v[x] = nn;

	} else if (op == 0x7) {
		/* add nn to v[x] */
		emu->v[x] += nn;

	} else if (op == 0x8 && n == 0x0) {
		/* store v[y] in v[x] */
		emu->v[x] = emu->v[y];

	} else if (op == 0x8 && n == 0x1) {
		/* v[x] = v[x] OR v[y] */
		emu->v[x] = emu->v[x] | emu->v[y];

	} else if (op == 0x8 && n == 0x2) {
		/* v[x] = v[x] AND v[y] */
		emu->v[x] = emu->v[x] & emu->v[y];

	} else if (op == 0x8 && n == 0x3) {
		/* v[x] = v[x] XOR v[y] */
		emu->v[x] = emu->v[x] ^ emu->v[y];

	} else if (op == 0x8 && n == 0x4) {
		/* v[x] = v[x] + v[y]
		   VF is 1 if carry, else 0 */
		uint16_t tmp = emu->v[x] + emu->v[y];
		emu->v[15] = (tmp & 0xFF00) ? 1 : 0;
		emu->v[x] = tmp;

	} else if (op == 0x8 && n == 0x5) {
		/* v[x] = v[x] - v[y]
		   VF is 0 if borrow, else 1 */
		uint16_t tmp = emu->v[x] - emu->v[y];
		emu->v[15] = (tmp & 0xFF00) ? 0 : 1;
		emu->v[x] = tmp;

	} else if (op == 0x8 && n == 0x6) {
		/* v[x] = v[y] SHR 1
		   VF is shifted bit */
		emu->v[15] = (emu->v[y] & 0x01) ? 1 : 0;
		emu->v[x] = emu->v[y] >> 1;

	} else if (op == 0x8 && n == 0x7) {
		/* v[x] = v[y] - v[x]
		   VF is 0 if borrow, else 1 */
		uint16_t tmp = emu->v[y] - emu->v[x];
		emu->v[15] = (tmp & 0xFF00) ? 0 : 1;
		emu->v[x] = tmp;

	} else if (op == 0x8 && n == 0xE) {
		/* v[x] = v[y] SHL 1
		   VF is shifted bit */
		emu->v[15] = (emu->v[y] & 0x80) ? 1 : 0;
		emu->v[x] = emu->v[y] << 1;

	} else if (op == 0x9) {
		/* skip if v[x] != v[y] */
		if (emu->v[x] != emu->v[y])
			emu->pc += 2;

	} else if (op == 0xA) {
		/* store nnn in i */
		emu->i = nnn;

	} else if (op == 0xB) {
		/* jump to v[0] + nnn */
		emu->pc = emu->v[0] + nnn;

	} else if (op == 0xC) {
		/* v[x] = random number with mask
		   use xorshift16 algorithm */
		emu->seed ^= emu->seed << 7;
		emu->seed ^= emu->seed >> 9;
		emu->seed ^= emu->seed << 8;
		emu->v[x] = emu->seed & nn;

	} else if (op == 0xD) {
		/* draw sprite at (x,y) with n bytes from address i
		   VF is 1 if pixels cleared, else 0 */
		emu->v[15] = draw_sprite(emu,
			emu->v[x] & 0x3F,
			emu->v[y] & 0x1F, n);

	} else if (op == 0xE && nn == 0x9E) {
		/* skip if key in v[x] is pressed */
		if (emu->cb.getkey(emu->v[x]))
			emu->pc += 2;

	} else if (op == 0xE && nn == 0xA1) {
		/* skip if key in v[x] is not pressed */
		if (!emu->cb.getkey(emu->v[x]))
			emu->pc += 2;

	} else if (op == 0xF && nn == 0x07) {
		/* store delay timer in v[x] */
		emu->v[x] = emu->dtim;

	} else if (op == 0xF && nn == 0x0A) {
		/* wait for key and store in v[x] */
		int ret = emu->cb.waitkey();
		if (ret < 0)
			return 16; /* EXIT: external event */
		emu->v[x] = ret;

	} else if (op == 0xF && nn == 0x15) {
		/* set delay timer to v[x] */
		emu->dtim = emu->v[x];

	} else if (op == 0xF && nn == 0x18) {
		/* set sound timer to v[x] */
		if (emu->cb.sound && !emu->stim && emu->v[x])
			emu->cb.sound(1);
		emu->stim = emu->v[x];

	} else if (op == 0xF && nn == 0x1E) {
		/* add v[x] to i */
		emu->i = (emu->i + emu->v[x]) & 0xFFF;

	} else if (op == 0xF && nn == 0x29) {
		/* set i to sprite address for v[x] */
		emu->i = 5 * emu->v[x];

	} else if (op == 0xF && nn == 0x33) {
		/* write v[x] in BCD to mem[i] */
		emu->mem[(emu->i+0) & 0xFFF] = (emu->v[x] / 100) % 10;
		emu->mem[(emu->i+1) & 0xFFF] = (emu->v[x] /  10) % 10;
		emu->mem[(emu->i+2) & 0xFFF] = (emu->v[x] /   1) % 10;

	} else if (op == 0xF && nn == 0x55) {
		/* store v[0]..v[x] to mem[i]
		   i = i + x + 1 */
TRACE("STORE ");
		for (int c = 0; c <= x; c++) {
TRACE(" V%x => %03x (%02x)", c, emu->i, emu->v[c]);
			emu->mem[emu->i & 0xFFF] = emu->v[c];
			emu->i = (emu->i + 1) & 0xFFF;
		}
TRACE("\n");

	} else if (op == 0xF && nn == 0x65) {
		/* fetch v[0]..v[x] from mem[i]
		   i = i + x + 1 */
TRACE("FETCH ");
		for (int c = 0; c <= x; c++) {
			emu->v[c] = emu->mem[emu->i & 0xFFF];
TRACE(" V%x <= %03x (%02x)", c, emu->i, emu->v[c]);
			emu->i = (emu->i + 1) & 0xFFF;
		}
TRACE("\n");

	} else {
		fprintf(stderr, "CHIP-8: invalid instruction %04x at %04x\n",
			instr, (emu->pc-2) & 0xFFF);
		return 2; /* EXIT: invalid instruction */
	}
	return 0;
}

/* execute program until exit or failure,
   1 => exit / return on empty stack
   2 => invalid instruction
   3 => stack overflow
  16 => external event (tick) */
int chip8_run(void* handle)
{
	struct emulator* emu = (struct emulator*)handle;

	/* TODO: better instruction loop */

	while(1) {
		/* execute a few instructions */
		for (int c = 0; c < SPEED; c++) {
			int ret = next_instr(emu);
			if (ret) return ret;
		}

		/* handle delay timer */
		if (emu->dtim) emu->dtim--;

		/* handle sound timer */
		if (emu->stim) {
			emu->stim--;
			if (!emu->stim && emu->cb.sound) {
				emu->cb.sound(0);
			}
		}

		/* tick the outside */
		if (emu->cb.tick())
			return 16;

		/* wait approximately 1/60s */
		usleep(16666);
	}

	return 0;
}
