#include <stdio.h>
#include <stdint.h>
#include <SDL2/SDL.h>

#include "sdl2_ui.h"

#define SCALE  8
#define BORDER 1
#define VOLUME 0x04

static uint32_t colors[2] = { 0x33333333, 0xCCCCCCCC };

static SDL_Window   *window;
static SDL_Renderer *renderer;
static SDL_Texture  *texture;

static SDL_AudioDeviceID audio = 0;

#define WIDTH  64
#define HEIGHT 32
static size_t fbsize = sizeof(uint32_t) *
	SCALE * (WIDTH + 2 * BORDER) *
	SCALE * (HEIGHT + 2 * BORDER);
static uint32_t* fb;

static int flag_quit = 0;
static int flag_beep = 0;

/* map 16-key keypad to modern keyboard */
struct {
	int code;
	int state;
	SDL_Keycode sdlcode;
} keylist[] = {
	/* numeric keypad, inverted layout */
	{ 0x1, 0, SDLK_KP_7,        },
	{ 0x2, 0, SDLK_KP_8,        },
	{ 0x3, 0, SDLK_KP_9,        },
	{ 0x4, 0, SDLK_KP_4,        },
	{ 0x5, 0, SDLK_KP_5,        },
	{ 0x6, 0, SDLK_KP_6,        },
	{ 0x7, 0, SDLK_KP_1,        },
	{ 0x8, 0, SDLK_KP_2,        },
	{ 0x9, 0, SDLK_KP_3,        },
	{ 0x0, 0, SDLK_KP_0,        },
	{ 0xA, 0, SDLK_KP_ENTER,    },
	{ 0xB, 0, SDLK_KP_PERIOD,   },
	{ 0xC, 0, SDLK_KP_DIVIDE,   },
	{ 0xD, 0, SDLK_KP_MULTIPLY, },
	{ 0xE, 0, SDLK_KP_MINUS,    },
	{ 0xF, 0, SDLK_KP_PLUS,     },
	/* arrow keys */
	{ 0x2, 0, SDLK_UP,    },
	{ 0x4, 0, SDLK_LEFT,  },
	{ 0x5, 0, SDLK_SPACE, },
	{ 0x6, 0, SDLK_RIGHT, },
	{ 0x8, 0, SDLK_DOWN,  },
	/* numbers and letters (a-f) */
	{ 0x0, 0, SDLK_0, },
	{ 0x1, 0, SDLK_1, },
	{ 0x2, 0, SDLK_2, },
	{ 0x3, 0, SDLK_3, },
	{ 0x4, 0, SDLK_4, },
	{ 0x5, 0, SDLK_5, },
	{ 0x6, 0, SDLK_6, },
	{ 0x7, 0, SDLK_7, },
	{ 0x8, 0, SDLK_8, },
	{ 0x9, 0, SDLK_9, },
	{ 0xA, 0, SDLK_a, },
	{ 0xB, 0, SDLK_b, },
	{ 0xC, 0, SDLK_c, },
	{ 0xD, 0, SDLK_d, },
	{ 0xE, 0, SDLK_e, },
	{ 0xF, 0, SDLK_f, },
};

/* audio callback */
static void audio_cb(void* userdata, uint8_t* data, int len)
{
	(void)userdata;

	if (flag_beep) {
		/* square wave */
		uint8_t d = 0;
		int c = 0;
		while(len--) {
			*data++ = d;
			if (c++ == 22050/440) { c = 0; d ^= VOLUME; }
		}
	} else {
		/* no audio */
		memset(data, 0, len);
	}
}

/* draw framebuffer to screen */
static void update_screen(void)
{
	void* pixels;
	int   pitch;
	SDL_LockTexture(texture, NULL, &pixels, &pitch);
	memcpy(pixels, fb, fbsize);
	SDL_UnlockTexture(texture);
	SDL_RenderCopy(renderer, texture, NULL, NULL);
	SDL_RenderPresent(renderer);
}

/* handle events */
static void handle_events(void)
{
	SDL_Event ev;
	while (SDL_PollEvent(&ev)) {
		switch (ev.type) {
		case SDL_QUIT:
			/* exit on window close */
			flag_quit = 1;
			return;

		case SDL_KEYDOWN:
			/* exit on escape key */
			if (ev.key.keysym.sym == SDLK_ESCAPE) {
				flag_quit = 1;
				return;
			}

			/* find key and set as pressed */
			for (unsigned int i = 0;
			     i < sizeof(keylist)/sizeof(keylist[0]); i++) {
				if (keylist[i].sdlcode == ev.key.keysym.sym) {
					keylist[i].state = 1; /* key pressed */
					break;
				}
			}
			break;

		case SDL_KEYUP:
			/* find key and set as not pressed */
			for (unsigned int i = 0;
			     i < sizeof(keylist)/sizeof(keylist[0]); i++) {
				if (keylist[i].sdlcode == ev.key.keysym.sym) {
					keylist[i].state = 0; /* not pressed */
					break;
				}
			}
			break;
		}
	}
}

/* initialize SDL2 engine */
int sdl2_init(const char* title, int nosound)
{
	SDL_AudioSpec want, get;

	/* initialize SDL video */
	if (SDL_Init(SDL_INIT_VIDEO)) goto fail;
	window = SDL_CreateWindow(title,
		SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED,
		SCALE * (WIDTH + 2 * BORDER), SCALE * (HEIGHT + 2 * BORDER), 0);
	if (!window) goto fail;
	renderer = SDL_CreateRenderer(window, -1,
		SDL_RENDERER_ACCELERATED | SDL_RENDERER_PRESENTVSYNC);
	if (!renderer) goto fail;
	texture = SDL_CreateTexture(renderer, SDL_PIXELFORMAT_ARGB8888,
		SDL_TEXTUREACCESS_STREAMING,
		SCALE * (WIDTH + 2 * BORDER), SCALE * (HEIGHT + 2 * BORDER));
	if (!texture) goto fail;

	/* allocate and clear framebuffer */
	fb = malloc(fbsize);
	if (!fb) goto fail;

	for(size_t i = 0; i < fbsize/4; i++)
		fb[i] = colors[0];

	flag_quit = 0;
	flag_beep = 0;

	if (!nosound) {
		/* initialize SDL audio and start playing */
		if (SDL_Init(SDL_INIT_AUDIO)) goto nosound;
		memset(&want, 0, sizeof(want));
		want.freq     = 22050;
		want.format   = AUDIO_S8;
		want.channels = 1;
		want.samples  = 512;
		want.callback = audio_cb;
		audio = SDL_OpenAudioDevice(NULL, 0, &want, &get, 0);
		if (!audio) goto nosound;
		SDL_PauseAudioDevice(audio, 0);
	}

	return 0;

nosound:
	fprintf(stderr, "SDL2: Audio initialization failed, disable sound.\n");
	return 0;

fail:
	if (fb)       free(fb);
	if (texture)  SDL_DestroyTexture(texture);
	if (renderer) SDL_DestroyRenderer(renderer);
	if (window)   SDL_DestroyWindow(window);
	return 1;
}

/* shutdown SDL2 engine */
void sdl2_destroy(void)
{
	flag_beep = 0;
	flag_quit = 0;
	if (audio)    SDL_CloseAudioDevice(audio);
	if (fb)       free(fb);
	if (texture)  SDL_DestroyTexture(texture);
	if (renderer) SDL_DestroyRenderer(renderer);
	if (window)   SDL_DestroyWindow(window);
}

/* get specific key state */
int sdl2_getkey(int key)
{
	/* find key and set as reported if pressed */
	for (unsigned int i = 0;
	     i < sizeof(keylist)/sizeof(keylist[0]); i++) {
		if (keylist[i].code == key && keylist[i].state == 1) {
			return 1;
		}
	}

	return 0;
}

/* wait for any keypress */
int sdl2_waitkey(void)
{
	update_screen();
	while(1) {
		/* handle events */
		handle_events();
		if (flag_quit)
			return -1;

		/* iterate keys */
		for (unsigned int i = 0;
		     i < sizeof(keylist)/sizeof(keylist[0]); i++) {
			if (keylist[i].state == 1) {
				/* only report key once */
				keylist[i].state = 2;
				return keylist[i].code;
			}
		}

		/* sleep a few ms */
		SDL_Delay(10);
	}
}

/* draw pixel */
void sdl2_draw(int x, int y, int c)
{
	if (x < 0 || y < 0 || x >= WIDTH || y >= HEIGHT)
		return;

	int pitch = SCALE * (WIDTH + 2 * BORDER);
	for (int yy = SCALE * (y + BORDER);
	     yy < SCALE * (y + BORDER + 1); yy++) {
		for (int xx = SCALE * (x + BORDER);
		     xx < SCALE * (x + BORDER + 1); xx++) {
			fb[(yy * pitch + xx)] = c ? colors[1] : colors[0];
		}
	}
}

/* enable or disable beep */
void sdl2_sound(int enable)
{
	if (audio) {
		flag_beep = enable;
	}
}

/* frame tick */
int sdl2_tick(void)
{
	handle_events();
	update_screen();
	return flag_quit;
}

